;;;; "rautil.scm", Scheme array accessory procedures.
;;; Copyright (c) 2000 Radey Shouman

;; Returns 0-origin array.
;; One SPEC should be supplied for each index of RA, a SPEC
;; may be:  an integer, taken as a value of the index
;;          #F, taken as all values of the index
;;       or a 2-list, (lower-bound upper-bound) of the index.
;;    Trailing SPECs may be elided, this is equivalent to specifying #F.
;;
;; Eg. (subarray '#2A((a b c) (1 2 3) (x y z)) '(1 2))
;;         => #2A((1 2 3) (x y z))
;;
;;     (subarray '#2A((a b c) (1 2 3) (x y z)) '(1 2) 1)
;;         => #1A(2 y)
;;
;;     (subarray '#2A((a b c) (1 2 3) (x y z)) 1)
;;         => #2A((1 2 3) (x y z))
;;
;;     (subarray '#2A((a b c) (1 2 3) (x y z)) 0)
;;         => #1A(a b c)

(define (subarray ra . specs)
  (let* ((oshape (array-shape ra))
	 (sps (let recur ((oshp oshape)
			  (sps specs))
		(cond ((null? oshp)
		       (or (null? sps)
			   (slib:error 'subarray 'rank-mismatch oshape specs))
		       '())
		      ((null? sps)
		       (cons (car oshp)
			     (recur (cdr oshp) '())))
		      ((not (car sps))
		       (cons (car oshp)
			     (recur (cdr oshp) (cdr sps))))
		      (else
		       (cons (car sps)
			     (recur (cdr oshp) (cdr sps))))))))
    (let loop ((oshp (reverse oshape))
	       (sps (reverse sps))
	       (nshp '())
	       (mapr (lambda () '())))
      (cond ((null? oshp)
	     (apply make-shared-array ra mapr nshp))
	    ((integer? (car sps))
	     (loop (cdr oshp) (cdr sps)
		   nshp
		   (lambda inds
		     (cons (car sps) (apply mapr inds)))))
	    ((pair? (car sps))
	     (or (<= (caar oshp) (caar sps) (cadar sps) (cadar oshp))
		 (slib:error 'subarray 'shape-mismatch oshape specs))
	     (let ((offset (caar sps)))
	       (loop (cdr oshp) (cdr sps)
		     (cons (+ 1 (- (cadar sps) (caar sps))) nshp)
		     (lambda (i . inds)
		       (cons (+ i offset) (apply mapr inds))))))))))

;; Return a subarray with shifted indices, each SHIFT is an exact
;; integer indicating how far to shift.
;; Eg.
;;  (array-shift '#(0 1 2 3 4) 1) => #1A(1 2 3 4) ;;shape is ((0 3))
;;  (array-shift '#(0 1 2 3 4) -1) => #1A(0 1 2 3) ;; shape is ((1 4))

;; Example
;;(require 'array-for-each)
;;(define (centered-difference ra)
;;  (array-map - (array-shift ra 1) (array-shift ra -1)))
;;(define (forward-difference ra)
;;  (array-map - (array-shift ra 1) ra))
;;(define (backward-difference ra)
;;  (array-map - ra (array-shift ra -1)))
;;
;;(centered-difference '#(0 1 3 5 9 22)) => #1A(3 4 6 17) ;;shape is ((1 4))
;;(backward-difference '#(0 1 3 5 9 22)) => #1A(1 2 2 4 13) ;; shape is ((1 5))
;;(forward-difference '#(0 1 3 5 9 22))  => #(1 2 2 4 13)  ;; shape is ((0 4))

(define (array-shift array . shifts)
  (let* ((shape (array-shape array))
	 (shifts (let recur ((shp shape)
			     (ss shifts))
		   (cond ((null? shp) '())
			 ((null? ss) (map (lambda (x) 0) shp))
			 ((integer? (car ss))
			  (cons (car ss) (recur (cdr shp) (cdr ss))))
			 (else
			  (error 'array-shift 'bad 'shift (car ss))))))
	 (nshp (map (lambda (spec shift)
		      (cond ((negative? shift)
			     (cons (- (car spec) shift) (cdr spec)))
			    ((positive? shift)
			     (list (car spec) (- (cadr spec) shift)))
			    (else spec)))
		    shape shifts)))
    (apply make-shared-array
	   array
	   (lambda inds (map + inds shifts))
	   nshp)))

;; Functional array mapping:  RA1 and RAS must be of the same rank,
;; except that some may be scalars.  The result array has a shape that
;; is the intersection of the shapes of all the argument arrays, ie it
;; is the largest array no larger than any argument array in any dimension.
(define (uniform-array-map prot fun ra1 . ras)
  (define (min-spec nspec spec)
    (cond ((number? nspec)
	   (min-spec (list 0 (- nspec 1)) spec))
	  ((number? spec)
	   (if (and (>= (car nspec) 0)
		    (< (cadr nspec) spec))
	       nspec
	       (list (max (car nspec) 0)
		     (min (cadr nspec) (- spec 1)))))
	  ((and (>= (car nspec) (car spec))
		(<= (cadr nspec) (cadr nspec)))
	   nspec)
	  (else (list (max (car nspec) (car spec))
		      (min (cadr nspec) (cadr spec))))))
  (define (min-shape ra1 ras)
    (let loop2 ((nshp (array-shape ra1))
		(ras ras))
      (if (null? ras)
	  nshp
	  (let ((shp (array-shape (car ras))))
	    (cond ((not shp) (loop2 nshp (cdr ras)))
		  ((not nshp) (loop2 shp (cdr ras)))
		  (else
		   (loop2 (map min-spec nshp shp)
			  (cdr ras))))))))
  (let ((dims (min-shape ra1 ras)))
    (if dims
	(let ((ret (dimensions->uniform-array dims prot)))
	  (apply array-map! ret fun ra1 ras)
	  ret)
	(apply fun ra1 ras))))

(define (array-map fun ra1 . ras)
  (apply uniform-array-map '() fun ra1 ras))

(define (uniform-array-index-map prot fun . shp)
  (let ((ret (dimensions->uniform-array shp prot)))
    (array-index-map! ret fun)
    ret))

;; Example:
;; (array-index-map identity 20))
;;       => #(0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19)

(define (array-index-map fun . shp)
  (apply uniform-array-index-map '() fun shp))
